
from typing import Iterator, Optional, Tuple
from logging import debug
from time import sleep
from socket import timeout

from eci.interfaces.primitives     import FourOctetAddress
from eci.interfaces.base_interface import BaseInterface
from eci.interfaces.messages       import DiscoveryResponseMsg, DiscoveryConfigMsg
from eci.interfaces.sockets        import DiscoverySendingSocket, DiscoveryReceivingSocket


IGNORE_REQUEST_SOURCE_ADDRESSES = set(['0.0.0.0'])


class DiscoveryInterface(BaseInterface):
    '''An interface to request and configure discovery information about the ECI.'''

    def __init__(self, *hosts: Tuple[str, ...]):
        debug(f'Initializing discovery interface on hosts="{hosts}"')
        super().__init__()
        self.hosts = hosts
        self._sending_sockets  = [DiscoverySendingSocket(host) for host in hosts] if len(hosts) else [DiscoverySendingSocket()]
        self._receiving_socket = DiscoveryReceivingSocket()

    def __enter__(self):
        super().__enter__()
        self._receiving_socket.__enter__()
        for socket in self._sending_sockets:
            socket.__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        for socket in self._sending_sockets:
            socket.__exit__()
        self._receiving_socket.__exit__()
        super().__exit__(*args, **kwargs)

    ################################################################

    def send_request(self) -> int:
        return [socket.send_request() for socket in self._sending_sockets]

    def send_config(self, msg: DiscoveryConfigMsg, addr: str) -> int:
        for socket in self._sending_sockets:
            r = socket.send_config(msg, str(FourOctetAddress(addr)))
        return r

    def recvfrom_disco_packet(self) -> Tuple[Optional[DiscoveryResponseMsg], Tuple[str, int]]:
        return self._receiving_socket.recvfrom_disco_packet()

    ################################################################

    def get_eci_list(self) -> Iterator[Tuple[Optional[DiscoveryResponseMsg], Tuple[str, int]]]:
        self.send_request()
        while True:
            sleep(1)
            try:
                msg, addr = msg, (host, port) = self.recvfrom_disco_packet()
                request_source_address = str(msg.request_source_address)
                if host not in self.hosts and request_source_address not in IGNORE_REQUEST_SOURCE_ADDRESSES:
                    if request_source_address in self.hosts or len(self.hosts) == 0:
                        yield msg, addr
            except (TimeoutError, timeout):
                break

    def forever_get_eci_list(self):
        while True:
            yield from self.get_eci_list()
